import zipfile as z
import tarfile as tar

from typing import Callable, Optional, Union

from .path import AbsolutePath, Path


ArFile = Union[z.ZipFile, tar.TarFile]
AbsolutePathTransform = Callable[[AbsolutePath], Path]


class Zipper:
    def __init__(self, format: str):
        self.format = format

    def open(self, path: AbsolutePath) -> "Archive":
        return Archive(path, self)

    def exists(self, path: AbsolutePath) -> bool:
        if self.format in {"zip", "ZIP"}:
            return z.is_zipfile(path.to_str())
        elif self.format in {"tgz", "tar.gz"}:
            return tar.is_tarfile(path.to_str())
        elif self.format in {"txz", "tar.xz"}:
            return tar.is_tarfile(path.to_str())
        else:
            raise NotImplementedError("Unknown format")

    def _open(self, path: AbsolutePath) -> ArFile:
        if self.format in {"zip", "ZIP"}:
            return z.ZipFile(path.to_str(), "w")
        elif self.format in {"tgz", "tar.gz"}:
            return tar.open(path.to_str(), "w:gz")
        elif self.format in {"txz", "tar.xz"}:
            return tar.open(path.to_str(), "w:xz")
        else:
            raise NotImplementedError("Unknown format")

    def add(
        self,
        path: AbsolutePath,
        target: ArFile,
        transform: Union[AbsolutePathTransform, None, str] = None,
    ) -> None:
        if isinstance(transform, str):
            arcname: Optional[str] = transform
        elif transform is not None:
            arcname = transform(path).to_str()
        else:
            arcname = None
        if arcname == "":
            arcname = "."
        if isinstance(target, z.ZipFile):
            target.write(path.to_str(), arcname=arcname)
        elif isinstance(target, tar.TarFile):
            target.add(path.to_str(), arcname=arcname)


class Archive:
    def __init__(self, path: AbsolutePath, zipper: Zipper):
        self.path = path
        self.zipper = zipper
        if path.exists():
            raise ValueError("File already exists. Append mode is not supported.")
        self.file: ArFile = zipper._open(path)
        self.default_transform: Union[AbsolutePathTransform, None, str] = None

    def use_transform(self, transform: AbsolutePathTransform):
        self.default_transform = transform

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def add(
        self,
        path: AbsolutePath,
        transform: Union[AbsolutePathTransform, None, str] = None,
    ):

        if transform is None and self.default_transform is not None:
            transform = self.default_transform

        self.zipper.add(path, self.file, transform=transform)

    def close(self):
        self.file.close()



from .path import Path


class Command:
    def __init__(self, obj):
        self.obj = obj

    def __call__(self, *args, **kwargs):
        return Command(self.obj(*args, **kwargs))

    def __getattr__(self, name):
        if name == "obj":
            return super().__getattribute__("obj")
        return Command(getattr(self.obj, name))

    def __repr__(self):
        if callable(self.obj):
            r = self.obj()
        else:
            r = self.obj

        if isinstance(r, Path):
            txt = str(r)
        elif hasattr(r, "__next__") or hasattr(r, "__iter__"):
            txt = "\n".join(map(str, r))
        else:
            txt = str(r)
        return txt



import sys

from typing import Union, Callable

from . import *
from .path import *
from .file import *
from . import ls as _ls
from . import find as _find
from . import cd as _cd
from .path import cwd as _cwd
from .command import Command
from .interpreter import Interpreter

ls = Command(_ls)
find = Command(_find)
cd = Command(_cd)
cwd = Command(_cwd)


def main():
    interp.interact()


def set_ps1(prompt: Union[str, Callable]):
    if callable(prompt):
        class P:
            def __str__(self):
                return prompt()
        sys.ps1 = P()
    else:
        sys.ps1 = prompt


interp = Interpreter(local=locals())


if __name__ == "__main__":
    main()



import re

from typing import Union, Optional, Iterable, Callable

from .path import AbsolutePath, RelativePath, Path, ensure_abs_path


PathPred = Callable[[AbsolutePath], bool]


class Query:
    def __init__(self, source: Iterable[Path], filter: Optional[PathPred] = None):
        self.source = source
        self._filter = filter

    def __iter__(self):
        for p in self.source:
            if isinstance(p, RelativePath):
                _p = AbsolutePath.from_rel(p)
            else:
                _p = p
            if self._filter is None or self._filter(_p):
                yield p

    def isfile(self):
        """Filter paths to keep only the files."""

        def isfile(p):
            return p.isfile()

        return Query(self, isfile)

    def isdir(self):
        """Filter the elements of the query to keep only the directories."""

        def isdir(p):
            return p.isdir()

        return Query(self, isdir)

    def name(self, pattern: Union[str, re.Pattern]):
        """Filter the elements of the query to keep only the ones matching `pattern`.

        :param pattern: a `str` or a `re.Pattern`, used as a regex to match the paths
        """
        if isinstance(pattern, re.Pattern):
            regex: re.Pattern = pattern
        else:
            regex = re.compile(pattern)

        def match(path: AbsolutePath):
            return bool(regex.fullmatch(path.base))

        return Query(self, match)

    def first(self):
        """Return the first element of the query."""
        return next(self.source)

    def last(self):
        """Return the last element of the query."""
        for it in self.source:
            pass
        return it

    def filter(self, pred: PathPred):
        """Create an arbitrary query filter from a predicate.

        :param pred: keep only the elements for which `pred(elem)` is true.
        """
        return Query(self, pred)

    def sort(self, key=None):
        """Sort the elements of the query.

        :param key: (optional) see :py:func:`sorted`
        """
        return Query(sorted(self, key=key))



import os
import os.path
import shutil
import shlex
import subprocess
import pathlib

from contextlib import contextmanager

from typing import Optional, Sequence, Iterator, Union, List

from .path import AbsolutePath, ensure_abs_path, SPath, RelativePath, cwd
from .zipper import Zipper
from .query import Query


def makeparam(default):
    def func(set=None):
        if set is None:
            return func.state
        else:
            func.state = set

    func.state = default
    return func


debug = makeparam(False)


def find(path: SPath = None, relative=False) -> Query:
    """Create a `Query` to recursively look for path under `path`.

    :param path: (optional, `cwd()`) a `str` or a `Path` target
    :param relative: (optional, False) wether to yield `RelativePath` or `AbsolutePath`
    """

    def walker() -> Iterator[Union[AbsolutePath, RelativePath]]:
        if path is None:
            _path = cwd()
        else:
            _path = ensure_abs_path(path)

        for dir, _, files in os.walk(_path.to_str()):
            if dir.startswith("/"):
                d = AbsolutePath.from_str(dir)
            else:
                d = RelativePath.from_str(dir).at(_path)
            if relative:
                _d = d.relative_to(_path)
                yield _d
                yield from (_d.add(file) for file in files)
            else:
                yield d
                yield from (d.add(file) for file in files)

    return Query(walker())


def ls(path: Optional[SPath] = None, relative: bool = True) -> Query:
    """Create a `Query` on the direct content of `path`

    :param path: (optional, `cwd()`) a `str` or a `Path` target
    :param relative: (optional, True) wether to yield `RelativePath` or `AbsolutePath`
    """

    def walker() -> Iterator[Union[AbsolutePath, RelativePath]]:
        if path is None:
            _path = cwd()
        else:
            _path = ensure_abs_path(path)

        if relative:
            yield from map(RelativePath, os.listdir(_path.to_str()))
        else:
            for name in os.listdir(_path.to_str()):
                yield _path.add(name)

    return Query(walker())


def mv(*args: SPath, force=False):
    """Move a set of source to a single target (like shell `mv`).

    :param *args: Path or str, the last one is the target
    """
    if len(args) == 1:
        raise ValueError("mv needs at least two parameters.")
    else:
        dest, *srcs = reversed(list(map(ensure_abs_path, args)))
        if dest.isdir():
            for src in srcs:
                b = src.base
                fdest = dest.add(b)
                if not force and fdest.exists():
                    raise ValueError("File already exists.")

            for src in srcs:
                b = src.base
                fdest = dest.add(b)
                if debug():
                    print(f"mv {src} {fdest}")
                else:
                    os.replace(src.to_str(), fdest.to_str())
        else:
            if len(srcs) > 1:
                raise ValueError(
                    "mv cannot move more than one source" " to the same dest"
                )
            if not force and dest.exists():
                raise ValueError("File already exists.")
            src = srcs[0]
            if debug():
                print(f"mv {src} {fdest}")
            else:
                os.replace(src.to_str(), dest.to_str())


def cp(*args: SPath, force=False):
    """Copy a set of source to a target (like shell `cp`).

    :param *args: Path or str, the last one is the target
    """
    if len(args) <= 1:
        raise ValueError("cp needs at least two parameters.")
    else:
        dest, *srcs = reversed(list(map(ensure_abs_path, args)))
        if dest.isdir():
            for src in srcs:
                b = src.base
                fdest = dest.add(b)
                if not force and fdest.exists():
                    raise ValueError(f"File already {fdest} exists.")

            for src in srcs:
                b = src.base
                fdest = dest.add(b)
                if debug():
                    print(f"cp -r {src} {fdest}")
                elif src.isdir():
                    shutil.copytree(src.to_str(), fdest.to_str())
                else:
                    shutil.copy(src.to_str(), fdest.to_str())
        else:
            if len(srcs) > 1:
                raise ValueError("cp cannot move more than one source to the same dest")
            if not force and dest.exists():
                raise ValueError("File already exists.")
            src = srcs[0]
            if debug():
                print(f"cp -r {src} {dest}")
            elif src.isdir():
                shutil.copytree(src.to_str(), dest.to_str())
            else:
                shutil.copy(src.to_str(), dest.to_str())


def rm(*args: SPath, recursive=False, force=False):
    if recursive:
        for p in map(ensure_abs_path, args):
            shutil.rmtree(p.to_str(), ignore_errors=force)
    else:
        if force:
            for p in map(ensure_abs_path, args):
                try:
                    if debug():
                        print(f"rm -rf {p}")
                    else:
                        os.remove(p.to_str())
                except Exception:
                    pass
        else:
            for p in map(ensure_abs_path, args):
                if debug():
                    print(f"rm -r {p}")
                else:
                    os.remove(p.to_str())


def mkdir(path: SPath, exist_ok=True):
    """Create a leaf directory and all intermedite ones.

    :param path: Target directory
    :param exist_ok: if False and path already exists, raise a `FileExistsError`
    """
    if debug():
        print(f"mkdir -p {path}")
    else:
        os.makedirs(ensure_abs_path(path).to_str(), exist_ok=exist_ok)


def archive(dest: SPath, *args: SPath, transform=None, format="tar.gz"):
    """Create a new archive containing a set of files."""
    if debug():
        raise NotImplementedError("archive is not yet supported in debug mode.")

    z = Zipper(format)
    if args:
        with z.open(ensure_abs_path(dest)) as ar:
            for f in map(ensure_abs_path, args):
                ar.add(f, transform=transform)
    else:
        return z.open(ensure_abs_path(dest))


def cd(to: Optional[SPath] = None):
    """Change current working directory."""
    if to is None:
        to = AbsolutePath.from_str("~")
    path = ensure_abs_path(to)
    os.chdir(path.to_str())


@contextmanager
def in_dir(d: SPath = None):
    """Context manager to temporarily change current working directory."""
    current = cwd()
    cd(d)
    try:
        yield
    finally:
        cd(current)


def sh(command: Union[str, List[str]], path: Optional[SPath] = None):
    """Run a shell command."""
    if isinstance(command, str):
        cmd = command
    else:
        cmd = " ".join(map(shlex.quote, command))

    if path:
        cwd = ensure_abs_path(path).to_str()
    else:
        cwd = None

    return subprocess.call(cmd, shell=True, cwd=cwd)


def sh_with_stdout(command: Union[str, List[str]], path: Optional[SPath] = None):
    """Run a shell command."""
    if isinstance(command, str):
        cmd = command
    else:
        cmd = " ".join(map(shlex.quote, command))

    if path:
        cwd = ensure_abs_path(path).to_str()
    else:
        cwd = None

    return subprocess.run(
        cmd, shell=True, cwd=cwd, check=True, text=True, stdout=subprocess.PIPE
    ).stdout


def cat(*args, overwrite=False):
    """Similar to cat command.

    `cat(a, b, c, ..., dest)` copy the content of a, b, c, ... into dest.
    Note: As opposed to the command cat, `cat(dest)` create an empty file `dest`.

    :param *args: there should be at least one argument. The last one is always
    the destination file path. The others are the sources.
    :param overwrite: (optional, False) overwrite dest if True.
    """
    dest, *sources = reversed(args)
    sources = reversed(sources)

    if not sources:
        if overwrite:
            p = pathlib.Path(dest)
            p.unlink()
            p.touch()
        return

    if overwrite:
        shutil.copyfile(sources[0], dest)
        sources = sources[1:]

    # TODO use sendfile on posix for faster write
    with open(dest, "ab") as d:
        for src in sources:
            with open(src, "rb") as s:
                d.write(s.read())
